import typing as tp

from substrateinterface import KeypairType

from .base import Base


class Mainnet(Base):
    """
    Class made for interacting with Crust Network to store files via Web3 services.
    """

    def __init__(self, seed: tp.Optional[str] = None, crypto_type: int = KeypairType.SR25519):

        super().__init__(chain="mainnet", seed=seed, crypto_type=crypto_type)

    def get_balance(self) -> int:
        """
        Get your account balance.

        :return: Account balance, pCRUs.

        """

        return self.query("System", "Account", self.keypair.ss58_address).value["data"]["free"]

    def get_appx_store_price(self, file_size: int) -> int:
        """
        Calculate an approximate price of the file.

        :param file_size: File size, bytes.

        :return: Approximate 6 months store cost, pCRUs.
        """

        # Returns base size-independent fee, pCRUs
        base_fee = self.query("Market", "FileBaseFee").value
        # Returns pCRUs fee per MB
        size_fee = self.query("Market", "FileByteFee").value

        return round(base_fee + size_fee * file_size / (1024 ** 2))

    def get_replicas(self, ipfs_cid: str) -> int:
        """
        Get number of file replicas in Crust network.

        :param ipfs_cid: Stored file cid.

        :return: Number of replicas.

        """

        return self.query("Market", "FilesV2", ipfs_cid).value["reported_replica_count"]

    def add_renewal_pool_balance(self, ipfs_cid: str, amount: int) -> tp.Tuple[str, str]:
        """
        Add funds to a renewal pool balance.
            https://wiki.crust.network/docs/en/storageUserGuide#124-renew-the-file-pool-balance. DOESN'T WORK ON SHADOW,
            will be executed in Mainnet.

        :param ipfs_cid: Stored file cid.
        :param amount: Amount of funds to be added, pCRUs. Make sure to add at balance which is enough for at least 6
            months. Use `calculate_appx_store_price` to check.

        :return: transaction hash, block_num-event_idx.

        """

        return self.extrinsic("Market", "add_prepaid", dict(cid=ipfs_cid, amount=amount))

    def store_file(self, ipfs_cid: str, file_size: int, tips: int = 0) -> tp.Tuple[str, str]:
        """
        Store a file `already uploaded` to IPFS providing its cid and size.

        :param ipfs_cid: Uploaded file cid.
        :param file_size: Uploaded file size, bytes.
        :param tips: Tips for the file host, pCRUs.

        :return: transaction hash, block_num-event_idx
        """

        return self.extrinsic(
            "Market", "place_storage_order", dict(cid=ipfs_cid, reported_file_size=file_size, tips=tips, _memo="")
        )



"""
Decorators to open interface instances when necessary only.

"""

import substrateinterface as substrate

from functools import wraps
from websocket._exceptions import WebSocketConnectionClosedException

from .constants import (
    CRUST_SS_58_FORMAT,
    CRUST_TYPE_REGISTRY_PRESET,
    CRUST_MAINNET_ENDPOINT,
    CRUST_ROCKY_TESTNET_ENDPOINT,
)


def check_interface_opened(func):
    """
    Open substrate node connection each time needed.

    :param func: wrapped function.

    :return: wrapped function after augmentations.

    """

    @wraps(func)
    def wrapper(crust_interface, *args, **kwargs):
        """
        Wrap decorated function with interface opening if it was closed.

        :param crust_interface: Crust interface instance in a decorated function.
        :param args: Wrapped function args.
        :param kwargs: Wrapped function kwargs.

        """

        if not crust_interface.interface:
            open_interface(crust_interface)

        try:
            res = func(crust_interface, *args, **kwargs)
        except (BrokenPipeError, WebSocketConnectionClosedException):
            open_interface(crust_interface)
            res = func(crust_interface, *args, **kwargs)

        return res

    return wrapper


def open_interface(interface_instance):
    """
    (Re-)open a new node interface instance.

    :param interface_instance: Interface to (re-)open.

    """

    interface_instance.interface = substrate.SubstrateInterface(
        url=interface_instance.remote_ws,
        ss58_format=CRUST_SS_58_FORMAT,
        type_registry_preset=CRUST_TYPE_REGISTRY_PRESET
        if interface_instance.remote_ws in [CRUST_MAINNET_ENDPOINT, CRUST_ROCKY_TESTNET_ENDPOINT]
        else None
    )



CRUST_ROCKY_TESTNET_ENDPOINT = "wss://rpc-rocky.crust.network"
CRUST_MAINNET_ENDPOINT = "wss://rpc.crust.network"
CRUST_SHADOW_ENDPOINT = "wss://rpc-shadow.crust.network/"
CRUST_TYPE_REGISTRY_PRESET = "crust"
CRUST_SS_58_FORMAT = 66



"""
Tool for interacting with Crust Shadow Parachain.

"""

import typing as tp

from substrateinterface import KeypairType

from .base import Base


class Shadow(Base):
    """
    Class made for interacting with Crust Shadow Kusama Parachain to store files via Web3 services. Most of the queries
        and extrinsics should be performed in Crust Main net with Mainnet, since files via Shadow are actually
        stored in Crust Mainnet.
    """

    def __init__(self, seed: tp.Optional[str] = None, crypto_type: int = KeypairType.SR25519):

        super().__init__(chain="shadow", seed=seed, crypto_type=crypto_type)

    def get_balance(self) -> int:
        """
        Get your account balance.

        :return: Account balance, pCSMs.

        """

        return self.query("System", "Account", self.keypair.ss58_address).value["data"]["free"]

    def store_file(self, ipfs_cid: str, file_size: int) -> tp.Tuple[str, str]:
        """
        Store a file `already uploaded` to IPFS providing its cid and size.

        :param ipfs_cid: Uploaded file cid.
        :param file_size: Uploaded file size, bytes.

        :return: transaction hash, block_num-event_idx.

        """

        return self.extrinsic("Xstorage", "place_storage_order_through_parachain", dict(cid=ipfs_cid, size=file_size))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['crustinterface']

package_data = \
{'': ['*']}

install_requires = \
['substrate-interface>=1.4.0,<2.0']

setup_kwargs = {
    'name': 'crust-interface-patara',
    'version': '0.1.1',
    'description': 'A simple tool to interact with Crust Shadow and Crust Mainnet. Under development',
    'long_description': '# crust-file-uploader\n\nhttps://crust.network/\nhttps://apps.crust.network/\nhttps://wiki.crust.network/en\nhttps://polkadot.js.org/apps/?rpc=wss%3A%2F%2Frpc-shadow.crust.network%2F#/explorer\n\nThis is a simple tool to pin your files sing Crust Network or Crust Shadow.\n\n## Setup\n### Installation:\n```bash\npip3 install crust-interface-patara\n```\n\n## Features\n\nThe module is divided into `Mainnet` and `Shadow`\n\n`Mainnet` provides Crust interaction functionality to check user balance, calculate file storage price, placing\nfile storage order, add tokens to renewal pool and checking replicas count.\n\n```python\nimport time\nfrom crustinterface import Mainnet\n\nseed = "seed"\nmainnet = Mainnet(seed=seed)\n\n# get any IPFS CID and size\ncid, size =  "QmbJtyu82TQSHU52AzRMXBENZGQKYqPsmao9dPuTeorPui", 18  # <any way to get an IPFS CID and size. One may use ipfshttpclient2 from IPFS-Toolkit>\n\n# Check balance\nbalance = mainnet.get_balance()\nprint(balance)\n\n# Check price in Main net. Price in pCRUs\nprice = mainnet.get_appx_store_price(int(size))\nprint(price)\n\n# Store file in Mainnet for CRUs\nfile_stored = mainnet.store_file(cid, size)\nprint(file_stored)\n\n# Add renewal pool\nfile_prepaid = mainnet.add_renewal_pool_balance(cid, price*2)\nprint(file_prepaid)\n\n\n# Get replicas\ntime.sleep(10)\nreplicas = mainnet.get_replicas(cid)\nprint(replicas)\n\n```\n\n`Shadow` allows you to perform `Xstorage` extrinsic in Crust Shadow network.\n```python\nfrom crustinterface import Shadow\nseed = "seed"\nshadow = Shadow(seed=seed)\n\n# get any IPFS CID and size\ncid, size =  "QmbJtyu82TQSHU52AzRMXBENZGQKYqPsmao9dPuTeorPui", 18  # <any way to get an IPFS CID and size. One may use ipfshttpclient2 from IPFS-Toolkit>\n\nprint(cid, size)\n\n# Check balance\nbalance = shadow.get_balance()\nprint(balance)\n\n# Store file in Shadow for CSMs\nfile_stored = shadow.store_file(cid, size)\nprint(file_stored)\n```\n',
    'author': 'PaTara43',
    'author_email': 'p040399@outlook.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/PaTara43/crust-interface-patara',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import typing as tp

from substrateinterface import SubstrateInterface, Keypair, KeypairType

from .constants import CRUST_SS_58_FORMAT, CRUST_MAINNET_ENDPOINT, CRUST_SHADOW_ENDPOINT, CRUST_ROCKY_TESTNET_ENDPOINT
from .decorators import check_interface_opened


class Base:
    """
    Base class with basic functionality for Crust interfaces
    """

    def __init__(self, chain: str = "shadow", seed: tp.Optional[str] = None, crypto_type: int = KeypairType.SR25519):
        """
        Accept basic information for interacting with Crust interfaces.

        :param chain: Chain to work with: shadow, mainnet or rocky.
        :param seed: User seed (needed to sign storage transactions).
        :param crypto_type: Account KeypairType.

        """

        if seed:
            self.keypair: Keypair = self.create_keypair(seed, crypto_type)
        else:
            self.keypair = None

        if chain == "shadow":
            self.remote_ws = CRUST_SHADOW_ENDPOINT
        elif chain == "mainnet":
            self.remote_ws = CRUST_MAINNET_ENDPOINT
        elif chain == "rocky":
            self.remote_ws = CRUST_ROCKY_TESTNET_ENDPOINT
        else:
            raise ValueError("Invalid chain name. Choose from [shadow, mainnet, rocky].")

        self.interface: tp.Optional[SubstrateInterface] = None

    @staticmethod
    def create_keypair(seed: str, crypto_type: int = KeypairType.SR25519) -> Keypair:
        """
        Create a keypair for further use.

        :param seed: Account seed (mnemonic or raw) as a key to sign transactions.
        :param crypto_type: Account crypto_type.

        :return: A Keypair instance used by substrate to sign transactions.

        """

        if seed.startswith("0x"):
            return Keypair.create_from_seed(
                seed_hex=hex(int(seed, 16)), ss58_format=CRUST_SS_58_FORMAT, crypto_type=crypto_type
            )
        else:
            return Keypair.create_from_mnemonic(seed, ss58_format=CRUST_SS_58_FORMAT, crypto_type=crypto_type)

    @check_interface_opened
    def extrinsic(self, call_module: str, call_function: str, params: tp.Optional[dict] = None) -> tp.Tuple[str, str]:
        """
        Submit an extrinsic in Crust Network.

        :param call_module: Call module.
        :param call_function: Call function.
        :param params: Call params.

        :return: transaction hash, block_num-event_idx.

        """

        call = self.interface.compose_call(call_module=call_module, call_function=call_function, call_params=params)
        signed_extrinsic = self.interface.create_signed_extrinsic(call=call, keypair=self.keypair)
        receipt = self.interface.submit_extrinsic(signed_extrinsic, wait_for_finalization=True)
        block_num: int = self.interface.get_block_number(receipt.block_hash)

        return receipt.extrinsic_hash, f"{block_num}-{receipt.extrinsic_idx}"

    @check_interface_opened
    def query(
        self, storage_module: str, storage_function: str, params: tp.Union[list, str, int, None] = None
    ) -> tp.Any:
        """
        Perform a query to a Crust Network.

        :param storage_module: Storage module.
        :param storage_function: Storage function.
        :param params: Query params.

        :return: Query result.

        """

        return self.interface.query(storage_module, storage_function, [params] if params else None)
